// ignore_for_file: lines_longer_than_80_chars
/// Documentation of the dart_pre_commit binary
///
/// You can run this script via `dart pub run dart_pre_commit [options]`. It
/// will create an instance of [Hooks] and invoke it to perform the pre commit
/// hooks. Check the documentation of the [Hooks] class for more details on what
/// the sepecific hooks do.
///
/// In order to be able to configure how the hooks should be run, you can
/// specify command line arguments to the script. The following tables list all
/// available options, organized into the same groups as shown when running
/// `dart pub run dart_pre_commit --help`.
///
/// ### Parsing Options
///  Option                             | Default | Description
/// ------------------------------------|---------|-------------
/// `-i`, `--[no-]fix-imports`          | on      | Format and sort imports of staged files.
/// `-f`, `--[no-]format`               | on      | Format staged files with dart format.
/// `-a`, `--[no-]analyze`              | on      | Run dart analyze to find issue for the staged files.
/// `-p`, `--[no-]check-pull-up`        | off     | Check if direct dependencies in the pubspec.lock have higher versions then specified in pubspec.yaml and warn if that's the case.
/// `-c`, `--[no-]continue-on-rejected` | off     | Continue checks even if a task rejects a certain file. The whole hook will still exit with rejected, but only after all files have been processed.
///
/// ### Other
///  Option                           | Default             | Description
/// ----------------------------------|---------------------|-------------
/// `-d`, `--directory=<dir>`         | `Directory.current` | Set the directory to run this command in. By default, it will run in the current working directory.
/// `-e`, `--[no-]detailed-exit-code` | off                 | Instead of simply 0/1 as exit code for 'commit ok' or 'commit needs user intervention', output exit codes according to the full hook result (See [HookResult]).
/// `-l`, `--log-level=<level>`       | `info`              | Specify the logging level for task logs. This only affects log details of tasks, not the status update message. Can be any of [LogLevel].
/// `-v`, `--version`                 | -                   | Show the version of the dart_pre_commit package.
/// `-h`, `--help`                    | -                   | Show this help.
library dart_pre_commit_bin;

import 'dart:io';

import 'package:args/args.dart'; // ignore: import_of_legacy_library_into_null_safe
import 'package:dart_pre_commit/dart_pre_commit.dart';
import 'package:dart_pre_commit/src/hooks_provider.dart';
import 'package:riverpod/riverpod.dart'; // ignore: import_of_legacy_library_into_null_safe

/// @nodoc
void main(List<String> args) {
  _run(args).then((c) => exitCode = c);
}

Future<int> _run(List<String> args) async {
  final di = ProviderContainer();
  final parser = ArgParser()
    ..addSeparator('Parsing Options:')
    ..addFlag(
      'fix-imports',
      abbr: 'i',
      defaultsTo: true,
      help: 'Format and sort imports of staged files.',
    )
    ..addFlag(
      'format',
      abbr: 'f',
      defaultsTo: true,
      help: 'Format staged files with dart format.',
    )
    ..addFlag(
      'analyze',
      abbr: 'a',
      defaultsTo: true,
      help: 'Run dart analyze to find issue for the staged files.',
    )
    ..addFlag(
      'check-pull-up',
      abbr: 'p',
      help: 'Check if direct dependencies in the pubspec.lock have '
          'higher versions then specified in pubspec.yaml and warn if '
          'that´s the case.',
    )
    ..addFlag(
      'continue-on-rejected',
      abbr: 'c',
      help: 'Continue checks even if a task rejects a certain file. The whole '
          'hook will still exit with rejected, but only after all files have '
          'been processed.',
    )
    ..addSeparator('Other:')
    ..addOption(
      'directory',
      abbr: 'd',
      help: 'Set the directory to run this command in. By default, it will run '
          'in the current working directory.',
      valueHelp: 'dir',
    )
    ..addFlag(
      'detailed-exit-code',
      abbr: 'e',
      help: 'Instead of simply 0/1 as exit code for "commit ok" or "commit '
          'needs user intervention", output exit codes according to the full '
          'hook result (See HookResult).',
    )
    ..addOption(
      'log-level',
      abbr: 'l',
      allowed: LogLevel.values.map((e) => e.name),
      defaultsTo: LogLevel.info.name,
      help: 'Specify the logging level for task logs. This only affects log '
          'details of tasks, not the status update message. The levels are:',
      valueHelp: 'level',
      allowedHelp: {
        LogLevel.debug.name: 'Print all messages',
        LogLevel.info.name: 'Print informational messages',
        LogLevel.warn.name: 'Print warnings and errors only',
        LogLevel.error.name: 'Print errors only',
        LogLevel.except.name: 'Print exceptions only',
        LogLevel.nothing.name: 'Print nothing at all',
      },
    )
    ..addFlag(
      'version',
      abbr: 'v',
      negatable: false,
      help: 'Show the version of the dart_pre_commit package.',
    )
    ..addFlag(
      'help',
      abbr: 'h',
      negatable: false,
      help: 'Show this help.',
    );

  try {
    final options = parser.parse(args);
    if (options['help'] as bool) {
      stdout.write(parser.usage);
      return 0;
    }

    final dir = options['directory'] as String?;
    if (dir != null) {
      Directory.current = dir;
    }

    final hooks = await di.read(HooksProvider.hookProvider(HooksConfig(
      fixImports: options['fix-imports'] as bool,
      format: options['format'] as bool,
      analyze: options['analyze'] as bool,
      pullUpDependencies: options['check-pull-up'] as bool,
      continueOnRejected: options['continue-on-rejected'] as bool,
    )).future);
    hooks.logger.logLevel = LogLevelX.parse(options['log-level'] as String);

    final result = await hooks();
    if (options['detailed-exit-code'] as bool) {
      return result.index;
    } else {
      return result.isSuccess ? 0 : 1;
    }
  } on FormatException catch (e) {
    stderr
      ..writeln('${e.message}\n')
      ..write(parser.usage);
    return 2;
  } on Exception catch (e, s) {
    di.read(HooksProviderInternal.taskLoggerProvider).except(e, s);
    return 127;
  } finally {
    di.dispose();
  }
}
